#include <octave/config.h>
#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>
#include <octave/help.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include "oct-engine.h"    

// DEFUN_DLD and the macros that it depends on are defined in the
// files defun-dld.h, defun.h, and defun-int.h.
DEFUN_DLD (mxrcv, args, nargout,
  "val = mxrcv(msgid, [mtype])\n\
\n\
Receive a scalar, vector or matrix val with mtype mtype from message\n\
queue msgid.\n\
If mtype=0, retrieve any mtype messages in the queue.\n\
If mtype>0, only retrieve the specific mtype messages.")
{    
    // The list of values to return.  See the declaration in oct-obj.h
    octave_value_list retval;

    // The arguments to this function are available in args.
    int nargin = args.length ();
    if (nargin<1 || nargin >2) {
	error ("need one or two arguments");
	return retval;
    }

    int msgid = (int)args(0).double_value();
    
    long mtype;
    if (nargin == 2) mtype =  (long)(args(1).double_value());
    else mtype = 0;

    mxbuf buf;
    if (msgrcv(msgid, (struct msgbuf *)&buf, sizeof(buf), mtype, 0)
	== -1) {
	error("mxrcv: can not msgrcv");
	return retval;
    }
    
    int nd = buf.nd;
    if (nd > 2) {
	error("rank > 2 is too big");
	return retval;
    }
    PyArray_TYPES type = buf.type;
    if (type != PyArray_DOUBLE && type != PyArray_CDOUBLE) {
	error("unsupported datatype");
	return retval;
    }
    MAJOR major = buf.major;
    long size = buf.size;
    int i, dims[2];
    for (i=0; i<nd; i++)
	dims[i] = buf.dims[i];
    if (i == 1) dims[1] = 1;
    octave_allocator allocator(size);
    char *data = (char *)allocator.alloc(size);
    
    int offset = 0;
    i = 0;
    while (1) {
	data[offset++] = *(buf.data+(i++));
	if (i == BUF_SIZE && offset != size) {
	    if (msgrcv(msgid, (struct msgbuf *)&buf, sizeof(buf),
		       mtype, 0) == -1) {
		allocator.free(data, size);
		error("mxsnd: can not msgsnd");
		return retval;
	    }
	    i = 0;
	}
	else if (offset == size) break;
    }

    if (type == PyArray_DOUBLE) {
	double *dp = (double *)data;
	Matrix m = Matrix(dims[0], dims[1]);
	for (int i=0; i<dims[0]; i++)
	    for (int j=0; j<dims[1]; j++)
		if (major == COLUMN)
		    m.elem(i,j) = *(dp+dims[0]*j+i);
		else
		    m.elem(i,j) = *(dp+dims[1]*i+j);
	retval(0) = octave_value(m);
    }
    else if (type == PyArray_CDOUBLE) {
	Complex *dp = (Complex *)data;
	ComplexMatrix m = ComplexMatrix(dims[0], dims[1]);
	for (int i=0; i<dims[0]; i++)
	    for (int j=0; j<dims[1]; j++)
		if (major == COLUMN)
		    m.elem(i,j) = *(dp+dims[0]*j+i);
		else
		    m.elem(i,j) = *(dp+dims[1]*i+j);
	retval(0) = octave_value(m);
    }

    allocator.free(data, size);
    return retval;
}





