/*
    PYMAT -- A Python module that exposes the MATLAB engine interface
             and supports passing NumPy arrays back and forth.

  Revision History
  ----------------

  Version 1.0 -- December 26, 1998, Andrew Sterian (asterian@umich.edu)
     * Initial release
*/
/*
   PYOCT -- A Python Module interface to Octave
   Based on PYMAT 1.0 by Andrew Sterian
   
   Version 0.1 -- 4/3/99, Min Xu
*/
#define PYOCT_VERSION "0.1"

#include <Python.h>
#include <arrayobject.h>
#define OCT_ENGINE
#include <oct-engine.h>

#define AND &&
#define OR  ||
#define EQ  ==
#define NEQ !=

static void pyGenericError(PyObject *pException, const char *fmt, ...)
{
    char *lBuf = new char[2*strlen(fmt) + 1024];

    va_list ap;
    va_start(ap, fmt);
    vsprintf(& lBuf[0], fmt, ap);
    va_end(ap);
    
    strcat(&lBuf[0], "\n");

    PyErr_SetString(pException, & lBuf[0]);
    delete [] lBuf;
}

#define pyassert(x,y) if (! (x)) { _pyassert(y); goto error_return; }

static void _pyassert(const char *pStr)
{
    PyErr_SetString(PyExc_RuntimeError, pStr);
}

//////////////////////////////////////////////////////////////////////////////

static char open_doc[] = 
"handle open([str]) : Opens a Octave engine session\n"
"\n"
"This function returns a handle to a new Octave engine session.\n"
"The optional 'str' parameter determines how Octave is started.\n"
"If empty or not specified, the session is started by executing\n"
"the simple command 'Octave'. Other options include specifying\n"
"a host name or a verbatim string to use to invoke the Octave\n"
"program. See the engOpen() documentation in the Octave Engine\n"
"reference for more information.\n";

static PyObject *pyoct_open(PyObject *, PyObject *args)
{
    Engine *ep;
    char *lStr = NULL;

    if (! PyArg_ParseTuple(args, "|s:open", &lStr)) return NULL;
    pyassert(sizeof(int) >= sizeof(Engine *), "Oops! Pointers on this architecture are too big to fit in integers");

    ep = engOpen(lStr);
    if (ep EQ NULL) {
        pyGenericError(PyExc_RuntimeError, "Unable to start Octave engine");
	return NULL;
    }

    return Py_BuildValue("i", int(ep));

error_return:
    Py_INCREF(Py_None);
    return Py_None;
}

static char close_doc[] = 
"void close(handle) : Closes Octave session\n"
"\n"
"This function closes the Octave session whose handle was returned\n"
"by a previous call to open(). Return 0 for success.\n";

static PyObject *pyoct_close(PyObject *, PyObject *args)
{
    int lHandle;
    
    if (! PyArg_ParseTuple(args, "i:close", &lHandle)) return 0;
    
    if (engClose((Engine *)lHandle) NEQ 0) {
        pyGenericError(PyExc_RuntimeError, "Unable to close session");
        return NULL;
    }

    return Py_BuildValue("i", 0);
}

static char eval_doc[] = 
"void eval(handle, string) : Evaluates string in Octave session\n"
"\n"
"This function evaluates the given string in the Octave session\n"
"associated with the handle. The handle is returned from a previous\n"
"call to open().\n"
"\n"
"Return 0 for success, NULL for failure.\n";

static PyObject *pyoct_eval(PyObject *, PyObject *args)
{
    char *lStr;
    int lHandle;
    
    if (! PyArg_ParseTuple(args, "is:eval", &lHandle, &lStr)) return 0;

    if (engEvalString((Engine *)lHandle, lStr) NEQ 0) {
        pyGenericError(PyExc_RuntimeError, "Unable to evaluate string in Octave workspace");
        return NULL;
    }

    return Py_BuildValue("i", 0);
}

static char get_doc[] = 
"array get(handle, name) : Gets a matrix from the Octave session\n"
"\n"
"This function extracts the matrix with the given name from a Octave\n"
"session associated with the handle. The handle is the return value from\n"
"a previous call to open(). The name parameter must be a string describing\n"
"the name of a matrix in the Octave workspace. On Win32 platforms,\n"
"only double-precision floating point arrays (real or complex) are supported.\n"
"1-D character strings are supported on UNIX platforms.\n"
"\n"
"Only 1-dimensional and 2-dimensional arrays are currently supported. Cell\n"
"arrays, structure arrays, etc. are not yet supported.\n"
"\n"
"The return value is a NumPy array with the same shape and elements as the\n"
"Octave array.\n";

static PyObject *pyoct_get(PyObject *, PyObject *args)
{
    char *lName;
    int lHandle;
    PyArrayObject *lArray;
    PyObject *out;

    if (! PyArg_ParseTuple(args, "is:get", &lHandle, &lName)) return 0;

    lArray = engGetArray((Engine *)lHandle, lName);
    if (lArray EQ 0) {
        pyGenericError(PyExc_RuntimeError, "Unable to get matrix from Octave workspace");
	return 0;
    }
    double *p = (double *)lArray->data;
    printf("Pass 1: nd=%d, dims=%d,%d data[0]=%f\n", lArray->nd,
	   lArray->dimensions[0], lArray->dimensions[1], p[0]);
    out = (PyObject *)lArray;
    printf("Pass 2:\n");
    Py_DECREF(lArray);
    printf("Pass 3:\n");
    Py_INCREF(out);
    printf("Pass 4:\n");
    return out;
}

static char put_doc[] = 
"void put(handle, name, array) : Places a matrix into the Octave session\n"
"\n"
"This function places the given array into a Octave workspace under the\n"
"name given with the 'name' parameter (which must be a string). The handle\n"
"is a value previously obtained from a call to open().\n"
"\n"
"The 'array' parameter must be either a NumPy array, list, or tuple\n"
"containing real or complex numbers, or a string. The Octave array will\n"
"have the same shape and same values, after conversion to double-precision\n"
"floating point, if necessary. A string parameter is converted to a Octave\n"
"char-valued array.\n";

static PyObject * pyoct_put(PyObject *, PyObject *args)
{
    char *lName;
    int lHandle;
    PyArrayObject *lArray;

    if (! PyArg_ParseTuple(args, "isO:put", &lHandle, &lName, &lArray)) return 0;

    if (engPutArray((Engine *)lHandle, lName, lArray) NEQ 0) {
        pyGenericError(PyExc_RuntimeError, "Unable to put matrix into Octave workspace");
        return 0;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef pyoctMethods[] = {
    { "open",       pyoct_open,       1, open_doc },
    { "close",      pyoct_close,      1, close_doc },
    { "eval",       pyoct_eval,       1, eval_doc },
    { "get",        pyoct_get,        1, get_doc },
    { "put",        pyoct_put,        1, put_doc },
    { 0, 0}
};

extern "C" void initpyoct(void);

void initpyoct(void)
{
    PyObject *module = Py_InitModule4("pyoct",
				      pyoctMethods,
"PyMat -- Octave Engine Interface\n"
"\n"
"  open  - Open a Octave engine session\n"
"  close - Close a Octave engine session\n"
"  eval  - Evaluates a string in the Octave session\n"
"  get   - Gets a matrix from the Octave session\n"
"  put   - Places a matrix into the Octave session\n"
"\n"
"The NumPy Numeric package must be installed for this module to be used.\n"
,
				      0,
				      PYTHON_API_VERSION);

    /* This macro, defined in arrayobject.h, loads the Numeric API interface */
    import_array();

    PyObject *dict = PyModule_GetDict(module);
    PyObject *item = PyString_FromString(PYOCT_VERSION);
    PyDict_SetItemString(dict, "__version__", item);
    Py_XDECREF(item);
}

